@@ -45,6 +45,9 @@ enum {
         SWITCH_PAGE
 };
 
+#define FACE_ICON_SIZE 24
+#define DIALOG_TIMEOUT_MSEC 60000
+
 static void gs_lock_plug_class_init (GSLockPlugClass *klass);
 static void gs_lock_plug_init       (GSLockPlug      *plug);
 static void gs_lock_plug_finalize   (GObject         *object);
@@ -60,12 +63,15 @@ struct GSLockPlugPrivate
         GtkWidget   *password_entry;
         GtkWidget   *capslock_label;
         GtkWidget   *progress_bar;
+        GtkWidget   *user_treeview;
 
         GtkWidget   *ok_button;
         GtkWidget   *cancel_button;
         GtkWidget   *logout_button;
         GtkWidget   *switch_button;
 
+        FusaManager *fusa_manager;
+
         gboolean     caps_lock_on;
         gboolean     switch_enabled;
         gboolean     logout_enabled;
@@ -115,7 +121,7 @@ gs_lock_plug_style_set (GtkWidget *widget,
         plug = GS_LOCK_PLUG (widget);
 
         gtk_container_set_border_width (GTK_CONTAINER (plug->vbox), 24);
-        gtk_box_set_spacing (GTK_BOX (plug->vbox), 24);
+        gtk_box_set_spacing (GTK_BOX (plug->vbox), 12);
 
         gtk_container_set_border_width (GTK_CONTAINER (plug->action_area), 0);
         gtk_box_set_spacing (GTK_BOX (plug->action_area), 5);
@@ -125,14 +131,77 @@ static void
 manager_new_console_cb (FusaManager  *manager,
 			FusaDisplay  *display,
 			const GError *error,
-			GSLockPlug   *plug)
+			gpointer      data)
 {
+        GSLockPlug *plug = data;
         g_signal_emit (plug,
                        lock_plug_signals [RESPONSE],
                        0,
                        GS_LOCK_PLUG_RESPONSE_CANCEL);
 }
 
+static void
+do_user_switch (GSLockPlug  *plug,
+                FusaDisplay *display)
+{
+        GdkScreen *screen;
+
+        if (gtk_widget_has_screen (plug->priv->user_treeview)) {
+                screen = gtk_widget_get_screen (plug->priv->user_treeview);
+        } else {
+                screen = gdk_screen_get_default ();
+        }
+
+        if (display) {
+                fusa_manager_activate_display (plug->priv->fusa_manager, display, screen,
+                                               manager_new_console_cb, plug, NULL);
+                return;
+        }
+
+        fusa_manager_new_console (plug->priv->fusa_manager, screen,
+                                  manager_new_console_cb, plug, NULL);
+}
+
+enum {
+        DISPLAY_NAME_COLUMN,
+        NAME_COLUMN,
+        ACTIVE_COLUMN,
+        PIXBUF_COLUMN,
+        N_COLUMNS
+};
+
+static void
+switch_user_response (GSLockPlug *plug)
+{
+        FusaDisplay      *display = NULL;
+        FusaUser         *user;
+        GtkTreeSelection *selection;
+        GtkTreeModel     *model;
+        GtkTreeIter       iter;
+        GSList           *displays;
+        char             *name;
+
+        selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (plug->priv->user_treeview));
+        gtk_tree_selection_get_selected (selection,
+                                         &model,
+                                         &iter);
+        gtk_tree_model_get (model, &iter, NAME_COLUMN, &name, -1);
+        if (name
+            && strcmp (name, "__new_user") != 0
+            && strcmp (name, "__separator") != 0) {
+                user = fusa_manager_get_user (plug->priv->fusa_manager, name);
+                displays = fusa_user_get_displays (user);
+                if (displays) {
+                        /* FIXME: just pick the first one for now */
+                        display = displays->data;
+                }
+        }
+
+        g_free (name);
+
+        do_user_switch (plug, display);
+}
+
 static void
 gs_lock_plug_response (GSLockPlug *plug,
                        gint        response_id)
@@ -182,24 +251,8 @@ gs_lock_plug_response (GSLockPlug *plug,
                                                                          plug);
                         return;
                 } else {
-                        GdkScreen   *screen;
-                        FusaManager *dm;
-
-                        if (gtk_widget_has_screen (GTK_WIDGET (plug)))
-                                screen = gtk_widget_get_screen (GTK_WIDGET (plug));
-                        else
-                                screen = gdk_screen_get_default ();
-
-                        dm = fusa_manager_ref_default ();
-
-                        /* this is async */
-                        fusa_manager_new_console (dm,
-                                                  screen,
-                                                  (FusaManagerDisplayCallback)manager_new_console_cb,
-                                                  plug,
-                                                  NULL);
-                        g_object_unref (dm);
-
+                        /* switch user */
+                        switch_user_response (plug);
                         return;
                 }
         }
@@ -295,6 +348,20 @@ is_capslock_on (void)
         return (states & ShiftMask) != 0;
 }
 
+static void
+restart_monitor_progress (GSLockPlug *plug)
+{
+        if (plug->priv->idle_id > 0) {
+                g_source_remove (plug->priv->idle_id);
+                plug->priv->idle_id = 0;
+        }
+
+        g_get_current_time (&plug->priv->start_time);
+        plug->priv->idle_id = g_timeout_add (50,
+                                             (GSourceFunc)monitor_progress,
+                                             plug);
+}
+
 static void
 gs_lock_plug_show (GtkWidget *widget)
 {
@@ -304,10 +371,7 @@ gs_lock_plug_show (GtkWidget *widget)
 
         capslock_update (GS_LOCK_PLUG (widget), is_capslock_on ());
 
-        g_get_current_time (&(GS_LOCK_PLUG (widget)->priv->start_time));
-        GS_LOCK_PLUG (widget)->priv->idle_id = g_timeout_add (50,
-                                                              (GSourceFunc)monitor_progress,
-                                                              GS_LOCK_PLUG (widget));
+        restart_monitor_progress (GS_LOCK_PLUG (widget));
 }
 
 static void
@@ -484,10 +548,9 @@ static GtkWidget *
 get_ok_button_for_page (gint page)
 {
         GtkWidget *align;
-        GtkWidget *widget;
+        GtkWidget *widget = NULL;
         GtkWidget *hbox;
-        const char *stock_id = NULL;
-        const char *label    = NULL;
+        const char *label = NULL;
 
         align = gtk_alignment_new (0.5, 0.5, 0, 0);
         hbox = gtk_hbox_new (FALSE, 2);
@@ -495,19 +558,20 @@ get_ok_button_for_page (gint page)
 
         switch (page) {
         case (AUTH_PAGE):
-                stock_id = GTK_STOCK_DIALOG_AUTHENTICATION;
                 label = N_("_Unlock");
+                widget = gtk_image_new_from_stock (GTK_STOCK_DIALOG_AUTHENTICATION,
+                                                   GTK_ICON_SIZE_BUTTON);
                 break;
         case (SWITCH_PAGE):
-                stock_id = GTK_STOCK_REFRESH;
-                label = N_("_New Login");
+                label = N_("_Switch User");
+                widget = gtk_image_new_from_icon_name ("stock_people",
+                                                       GTK_ICON_SIZE_BUTTON);
                 break;
         default:
                 g_assert ("Invalid notebook page");
                 break;
         }
 
-        widget = gtk_image_new_from_stock (stock_id, GTK_ICON_SIZE_BUTTON);
         gtk_box_pack_start (GTK_BOX (hbox), widget, FALSE, FALSE, 0);
 
         widget = gtk_label_new_with_mnemonic (label);
@@ -520,10 +584,9 @@ static GtkWidget *
 get_switch_button_for_page (gint page)
 {
         GtkWidget *align;
-        GtkWidget *widget;
+        GtkWidget *widget = NULL;
         GtkWidget *hbox;
-        const char *stock_id = NULL;
-        const char *label    = NULL;
+        const char *label = NULL;
 
         align = gtk_alignment_new (0.5, 0.5, 0, 0);
         hbox = gtk_hbox_new (FALSE, 2);
@@ -531,19 +594,20 @@ get_switch_button_for_page (gint page)
 
         switch (page) {
         case (AUTH_PAGE):
-                stock_id = GTK_STOCK_REFRESH;
-                label = N_("_New Login");
+                label = N_("_Switch User");
+                widget = gtk_image_new_from_icon_name ("stock_people",
+                                                       GTK_ICON_SIZE_BUTTON);
                 break;
         case (SWITCH_PAGE):
-                stock_id = GTK_STOCK_DIALOG_AUTHENTICATION;
                 label = N_("_Unlock");
+                widget = gtk_image_new_from_stock (GTK_STOCK_DIALOG_AUTHENTICATION,
+                                                   GTK_ICON_SIZE_BUTTON);
                 break;
         default:
                 g_assert ("Invalid notebook page");
                 break;
         }
 
-        widget = gtk_image_new_from_stock (stock_id, GTK_ICON_SIZE_BUTTON);
         gtk_box_pack_start (GTK_BOX (hbox), widget, FALSE, FALSE, 0);
 
         widget = gtk_label_new_with_mnemonic (label);
@@ -689,6 +753,8 @@ entry_key_press (GtkWidget   *widget,
 {
         gboolean capslock_on;
 
+        restart_monitor_progress (plug);
+
         capslock_on = is_capslock_on ();
 
         if (capslock_on != plug->priv->caps_lock_on)
@@ -838,6 +904,263 @@ gs_lock_plug_set_default_response (GSLockPlug *plug,
         g_list_free (children);
 }
 
+typedef struct
+{
+        GtkTreeIter iter;
+        GtkWidget  *tree;
+} DisplayChangedData;
+
+static void
+user_displays_changed_cb (FusaUser           *user,
+                          DisplayChangedData *data)
+{
+        const char   *name;
+        gboolean      is_active;
+        int           n_displays;
+        GdkPixbuf    *pixbuf;
+        int           icon_size = FACE_ICON_SIZE;
+        GtkTreeModel *filter_model;
+        GtkTreeModel *model;
+
+        name = fusa_user_get_user_name (user);
+        n_displays = fusa_user_get_n_displays (user);
+        is_active = n_displays > 0;
+        pixbuf = fusa_user_render_icon (user, data->tree, icon_size, is_active);
+
+        filter_model = gtk_tree_view_get_model (GTK_TREE_VIEW (data->tree));
+        model = gtk_tree_model_filter_get_model (GTK_TREE_MODEL_FILTER (filter_model));
+        g_message ("Disp changed");
+        gtk_list_store_set (GTK_LIST_STORE (model), &data->iter,
+                            NAME_COLUMN, name,
+                            DISPLAY_NAME_COLUMN, fusa_user_get_display_name (user),
+                            ACTIVE_COLUMN, is_active,
+                            PIXBUF_COLUMN, pixbuf,
+                            -1);
+}
+
+static void
+populate_model (GSLockPlug   *plug,
+                GtkListStore *store)
+{
+        GtkTreeIter   iter;
+        GSList       *users;
+        GdkPixbuf    *pixbuf;
+        int           icon_size = FACE_ICON_SIZE;
+        GtkIconTheme *theme;
+
+        if (gtk_widget_has_screen (plug->priv->user_treeview))
+                theme = gtk_icon_theme_get_for_screen (gtk_widget_get_screen (plug->priv->user_treeview));
+        else
+                theme = gtk_icon_theme_get_default ();
+        
+        pixbuf = gtk_icon_theme_load_icon (theme, "gdm", icon_size, 0, NULL);
+
+#if 0
+        gtk_list_store_append (store, &iter);
+        gtk_list_store_set (store, &iter,
+                            DISPLAY_NAME_COLUMN, _("Log in as a new user"),
+                            NAME_COLUMN, "__new_user",
+                            PIXBUF_COLUMN, pixbuf,
+                            -1);
+        gtk_list_store_append (store, &iter);
+        gtk_list_store_set (store, &iter,
+                            DISPLAY_NAME_COLUMN, NULL,
+                            NAME_COLUMN, "__separator",
+                            -1);
+#endif
+
+        users = fusa_manager_list_users (plug->priv->fusa_manager);
+
+        while (users) {
+                FusaUser           *user;
+                gboolean            is_active;
+                guint               n_displays;
+                DisplayChangedData *ddata;
+
+                user = users->data;
+
+                /* skip the current user */
+                if (fusa_user_get_uid (user) == getuid ()) {
+                        users = g_slist_delete_link (users, users);
+                        continue;
+                }
+
+                n_displays = fusa_user_get_n_displays (user);
+                is_active = n_displays > 0;
+
+                /* this requires the following to scale well:
+                   http://bugzilla.gnome.org/show_bug.cgi?id=310418 */
+                pixbuf = fusa_user_render_icon (user, plug->priv->user_treeview, icon_size, is_active);
+
+                gtk_list_store_append (store, &iter);
+                gtk_list_store_set (store, &iter,
+                                    NAME_COLUMN, fusa_user_get_user_name (user),
+                                    DISPLAY_NAME_COLUMN, fusa_user_get_display_name (user),
+                                    ACTIVE_COLUMN, is_active,
+                                    PIXBUF_COLUMN, pixbuf,
+                                    -1);
+
+                ddata = g_new0 (DisplayChangedData, 1);
+                ddata->iter = iter;
+                ddata->tree = plug->priv->user_treeview;
+
+                g_signal_connect_data (user, "displays-changed",
+                                       G_CALLBACK (user_displays_changed_cb), ddata,
+                                       (GClosureNotify) g_free, 0);
+
+                users = g_slist_delete_link (users, users);
+        }
+}
+
+static int
+compare_users (GtkTreeModel *model,
+               GtkTreeIter  *a,
+               GtkTreeIter  *b,
+               gpointer      user_data)
+{
+        char *name_a;
+        char *name_b;
+        char *label_a;
+        char *label_b;
+        int   result;
+
+        gtk_tree_model_get (model, a, NAME_COLUMN, &name_a, -1);
+        gtk_tree_model_get (model, b, NAME_COLUMN, &name_b, -1);
+        gtk_tree_model_get (model, a, DISPLAY_NAME_COLUMN, &label_a, -1);
+        gtk_tree_model_get (model, b, DISPLAY_NAME_COLUMN, &label_b, -1);
+
+        if (! name_a)
+                return 1;
+        else if (! name_b)
+                return -1;
+
+        if (strcmp (name_a, "__new_user") == 0)
+                return -1;
+        else if (strcmp (name_b, "__new_user") == 0)
+                return 1;
+        else if (strcmp (name_a, "__separator") == 0)
+                return -1;
+        else if (strcmp (name_b, "__separator") == 0)
+                return 1;
+
+        if (! label_a)
+                return 1;
+        else if (! label_b)
+                return -1;
+
+        result = strcmp (label_a, label_b);
+
+        g_free (label_a);
+        g_free (label_b);
+        g_free (name_a);
+        g_free (name_b);
+
+        return result;
+}
+
+static gboolean
+separator_func (GtkTreeModel *model,
+                GtkTreeIter  *iter,
+                gpointer      data)
+{
+        int   column = GPOINTER_TO_INT (data);
+        char *text;
+        
+        gtk_tree_model_get (model, iter, column, &text, -1);
+        
+        if (text && strcmp (text, "__separator") == 0)
+                return TRUE;
+        
+        g_free (text);
+
+        return FALSE;
+}
+
+static gboolean
+filter_out_users (GtkTreeModel *model,
+                  GtkTreeIter  *iter,
+                  GSLockPlug   *plug)
+{
+        gboolean is_active;
+        gboolean visible;
+        char    *name;
+
+        gtk_tree_model_get (model, iter,
+                            NAME_COLUMN, &name,
+                            ACTIVE_COLUMN, &is_active,
+                            -1);
+        if (! name)
+                return FALSE;
+
+        if (strcmp (name, "__new_user") == 0
+            || strcmp (name, "__separator") == 0) {
+                visible = TRUE;
+        } else {
+                /* FIXME: do we show all users or only active ones? */
+                visible = TRUE;
+                /*visible = is_active;*/
+        }
+
+        g_free (name);
+
+        return visible;
+}
+
+static void
+setup_treeview (GSLockPlug *plug)
+{
+        GtkListStore      *store;
+        GtkTreeViewColumn *column;
+        GtkCellRenderer   *renderer;
+        GtkTreeModel      *filter;
+
+        store = gtk_list_store_new (N_COLUMNS,
+                                    G_TYPE_STRING,
+                                    G_TYPE_STRING,
+                                    G_TYPE_BOOLEAN,
+                                    GDK_TYPE_PIXBUF);
+        populate_model (plug, store);
+
+        filter = gtk_tree_model_filter_new (GTK_TREE_MODEL (store), NULL);
+
+        gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER (filter),
+                                                (GtkTreeModelFilterVisibleFunc) filter_out_users,
+                                                plug,
+                                                NULL);
+
+        gtk_tree_view_set_model (GTK_TREE_VIEW (plug->priv->user_treeview),
+                                 filter);
+
+        g_object_unref (store);
+        g_object_unref (filter);
+
+        renderer = gtk_cell_renderer_pixbuf_new ();
+        column = gtk_tree_view_column_new_with_attributes ("Image", renderer,
+                                                           "pixbuf", PIXBUF_COLUMN,
+                                                           NULL);
+        gtk_tree_view_append_column (GTK_TREE_VIEW (plug->priv->user_treeview), column);
+
+        renderer = gtk_cell_renderer_text_new ();
+        column = gtk_tree_view_column_new_with_attributes ("Name", renderer,
+                                                           "text", DISPLAY_NAME_COLUMN,
+                                                           NULL);
+        gtk_tree_view_append_column (GTK_TREE_VIEW (plug->priv->user_treeview), column);
+
+        gtk_tree_view_set_row_separator_func (GTK_TREE_VIEW (plug->priv->user_treeview),
+                                              separator_func,
+                                              GINT_TO_POINTER (NAME_COLUMN),
+                                              NULL);
+
+        gtk_tree_view_column_set_sort_column_id (column, 0);
+        gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (store),
+                                         0,
+                                         compare_users,
+                                         NULL, NULL);
+        gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (store),
+                                              0,
+                                              GTK_SORT_ASCENDING);
+}
+
 static void
 gs_lock_plug_init (GSLockPlug *plug)
 {
@@ -854,6 +1177,8 @@ gs_lock_plug_init (GSLockPlug *plug)
 
         plug->priv = GS_LOCK_PLUG_GET_PRIVATE (plug);
 
+        plug->priv->fusa_manager = fusa_manager_ref_default ();
+
         /* Dialog emulation */
 
         plug->vbox = gtk_vbox_new (FALSE, 0);
@@ -957,14 +1282,14 @@ gs_lock_plug_init (GSLockPlug *plug)
         hbox = gtk_hbox_new (FALSE, 12);
         gtk_notebook_append_page (GTK_NOTEBOOK (plug->priv->notebook), hbox, NULL);
 
-        widget = gtk_image_new_from_stock (GTK_STOCK_REFRESH, GTK_ICON_SIZE_DIALOG);
+        widget = gtk_image_new_from_icon_name ("stock_people", GTK_ICON_SIZE_DIALOG);
         gtk_misc_set_alignment (GTK_MISC (widget), 0.5, 0);
         gtk_box_pack_start (GTK_BOX (hbox), widget, FALSE, FALSE, 0);
 
         vbox = gtk_vbox_new (FALSE, 12);
         gtk_box_pack_start (GTK_BOX (hbox), vbox, TRUE, TRUE, 0);
 
-        widget = gtk_label_new (_("Log in as another user?"));
+        widget = gtk_label_new (_("Switch to another user?"));
         gtk_misc_set_alignment (GTK_MISC (widget), 0, 0.5);
         gtk_box_pack_start (GTK_BOX (vbox), widget, FALSE, FALSE, 0);
 
@@ -982,10 +1307,19 @@ gs_lock_plug_init (GSLockPlug *plug)
 
         gtk_label_set_attributes (GTK_LABEL (widget), pattrlist);
 
-        widget = gtk_label_new (_("This option will bring you to the login screen.\n"
-                                  "From the login screen you may log in to this system\n"
-                                  "as another user or select 'Quit' to return to this screen."));
-        gtk_box_pack_start (GTK_BOX (vbox), widget, FALSE, FALSE, 0);
+        widget = gtk_scrolled_window_new (NULL, NULL);
+        gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (widget),
+                                             GTK_SHADOW_IN);
+        gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (widget),
+                                        GTK_POLICY_NEVER,
+                                        GTK_POLICY_AUTOMATIC);
+        gtk_box_pack_start (GTK_BOX (vbox), widget, TRUE, TRUE, 0);
+
+        plug->priv->user_treeview = gtk_tree_view_new ();
+        gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (plug->priv->user_treeview), FALSE);
+        gtk_container_add (GTK_CONTAINER (widget), plug->priv->user_treeview);
+
+        setup_treeview (plug);
 
         /* Progress bar */
 
@@ -1052,7 +1386,7 @@ gs_lock_plug_init (GSLockPlug *plug)
         if (! plug->priv->switch_enabled)
                 gtk_widget_hide (plug->priv->switch_button);
 
-        plug->priv->timeout = 30000;
+        plug->priv->timeout = DIALOG_TIMEOUT_MSEC;
 
         g_signal_connect (plug->priv->switch_button, "clicked",
                           G_CALLBACK (switch_page), plug);
@@ -1103,6 +1437,8 @@ gs_lock_plug_finalize (GObject *object)
 
         g_return_if_fail (plug->priv != NULL);
 
+        g_object_unref (plug->priv->fusa_manager);
+
         if (plug->priv->password_check_idle_id > 0) {
                 g_source_remove (plug->priv->password_check_idle_id);
                 plug->priv->password_check_idle_id = 0;
