@@ -284,30 +284,35 @@ spawn_on_window (GSWindow *window,
 {
         int         argc;
         char      **argv;
-        char       *envp [5];
-        int         nenv = 0;
-        int         i;
+        char       *str;
+        GPtrArray  *env;
         gboolean    result;
         GIOChannel *channel;
         int         standard_output;
         int         child_pid;
         int         id;
+        int         i;
 
         if (!g_shell_parse_argv (command, &argc, &argv, NULL))
                 return FALSE;
 
-        envp[nenv++] = g_strdup_printf ("DISPLAY=%s",
-                                        gdk_display_get_name (gdk_display_get_default ()));
-        envp[nenv++] = g_strdup_printf ("HOME=%s",
-                                        g_get_home_dir ());
-        envp[nenv++] = g_strdup_printf ("PATH=%s", g_getenv ("PATH"));
-        envp[nenv++] = g_strdup_printf ("SESSION_MANAGER=%s", g_getenv ("SESSION_MANAGER"));
-        envp[nenv++] = NULL;
+        env = g_ptr_array_new ();
+
+        str = gdk_screen_make_display_name (GTK_WINDOW (window)->screen);
+        g_ptr_array_add (env, g_strdup_printf ("DISPLAY=%s", str));
+        g_free (str);
+
+        g_ptr_array_add (env, g_strdup_printf ("HOME=%s",
+                                               g_get_home_dir ()));
+        g_ptr_array_add (env, g_strdup_printf ("PATH=%s", g_getenv ("PATH")));
+        g_ptr_array_add (env, g_strdup_printf ("SESSION_MANAGER=%s", g_getenv ("SESSION_MANAGER")));
+        g_ptr_array_add (env, g_strdup_printf ("XAUTHORITY=%s", g_getenv ("XAUTHORITY")));
+        g_ptr_array_add (env, NULL);
 
         result = gdk_spawn_on_screen_with_pipes (GTK_WINDOW (window)->screen,
                                                  g_get_home_dir (),
                                                  argv,
-                                                 envp,
+                                                 (char **)env->pdata,
                                                  G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_SEARCH_PATH,
                                                  NULL,
                                                  NULL,
@@ -317,6 +322,10 @@ spawn_on_window (GSWindow *window,
                                                  NULL,
                                                  NULL);
 
+        for (i = 0; i < env->len; i++)
+                g_free (g_ptr_array_index (env, i));
+        g_ptr_array_free (env, TRUE);
+
         if (!result)
                 return FALSE;
 
@@ -339,9 +348,6 @@ spawn_on_window (GSWindow *window,
 
         g_io_channel_unref (channel);
 
-        for (i = 0; i < nenv; i++)
-                g_free (envp [i]);
-
         g_strfreev (argv);
 
         return result;
